function [D, X, sigmas] = dl_gauss_l1_sig(Y, D, sigmas, lam, gam, niter, nit_sig, met_adapt, ...
                                          nit_fista, step_fista)
% Input:
%    Y          - matrix of training signals (on columns)
%    D          - dictionary, central atoms
%    sigmas     - vector of standard deviations associated with the central atoms
%    lam        - weight of the error term
%    gam        - weight of the 1-norm term
%    niter      - number of iterations
%    nit_sig    - number of iterations at which sigmas are updated
%    met_adapt  - adaptation method; 0: number of nonzero coefficients,
%                                    1: sum of absolute values of coefficients
%                                    2: 2-norm of the coefficient vector
%    nit_fista  - number of iterations for FISTA
%    step_fista - step size in FISTA
% Output:
%    D          - learned dictionary
%    X          - representation matrix
%    sigmas     - adapted standard deviations (a permutation of the initial values?)

% BD 13.01.2024

[m,n] = size(D);
N = size(Y,2);
if length(sigmas) == 1   % only for backward compatibility (sigmas shouldn't be equal)
  sigmas = sigmas * ones(n,1);
else
  sigmas = sigmas(:);
end

av_sigma = mean(sigmas);
n_change_sig = niter / nit_sig;  % number of changes of the sigmas
sigmas_init = sort(sigmas,'descend');
delta_sig = (sigmas_init - av_sigma) / n_change_sig;
sigmas = av_sigma * ones(n,1);

X = zeros(n,N);

for it = 1 : niter
  %s2 = 1 ./ sigmas ./ sigmas;
  sig2 = sigmas .* sigmas;
  Dnew = zeros(m,n);
  count_atoms = zeros(n,1);
  weight_atoms = zeros(n,1);
  norm2_w = zeros(n,1);
  for iN = 1 : N
    y = Y(:,iN);
    x = X(:,iN);
    x = my_fista(y, D, gam/lam, step_fista, x, nit_fista);

    r = y - D*x;     % residual
    Da = D;

    % update actual atoms one by one, in random order
    p = randperm(n);
    for j = p
      d = Da(:,j);
      r = r + d*x(j);  % remove contribution of current atom
      % update atom and normalize
      %Da(:,j) = (1/(lam*x(j)*x(j) + s2(j))) * (lam*x(j)*r + s2(j)*D(:,j));
      d = (1/(lam*x(j)*x(j)*sig2(j) + 1)) * (lam*sig2(j)*x(j)*r + D(:,j));
      %d = d/norm(d);
      % update representation coefficient
      x(j) = my_wthresh(d'*r, gam/lam/2);
      % update residual
      r = r - d*x(j);
      Da(:,j) = d;
    end
    Da = normc(Da); % atom normalization
    X(:,iN) = x;

    support = find(x~=0);
    count_atoms(support) = count_atoms(support)+1;
    weight_atoms = weight_atoms + abs(x);
    norm2_w = norm2_w + x.*x;
    Dnew(:,support) = Dnew(:,support) + Da(:,support);
  end

  Dnew = normc(Dnew);
  norm(D-Dnew);
  D = Dnew;

  % update sigmas (by use; other ideas?)
  if rem(it, nit_sig) == 0
    sigmas = av_sigma + (it/nit_sig)*delta_sig; % sigmas sorted decreasingly
    switch met_adapt
      case 0
        [~,Ip] = sort(count_atoms,'descend');
      case 1
        [~,Ip] = sort(weight_atoms,'descend');
      case 2
        [~,Ip] = sort(sqrt(norm2_w),'descend');
    end
    sigmas(Ip) = sigmas;
  end
end

%norm(sort(sigmas,'descend') - sigmas_init)

%---------------------------------------------
function x = my_fista(y, D, lam, t, x, nit)

x1 = x;
s1 = 1;
for it = 1 : nit
  x2 = x1;
  s2 = s1;
  x1 = my_wthresh(x - 2*t*D'*(D*x - y), t*lam);
  s1 = (1 + sqrt(1+4*s2*s2))/2;
  x = x1 + ((s2-1)/s1)*(x1-x2);
end

%---------------------------------------------
function y = my_wthresh(x, t)

tmp = (abs(x)-t);
tmp = (tmp+abs(tmp))/2;
y   = sign(x).*tmp;
