% Main test program for Gaussian DL algorithms on ADBench datasets.
% Programs for paper "Atom dimension adaptation for infinite set dictionary learning",
% by A. Baltoiu, D.C. Ilie-Ablachim, B.Dumitrescu.
% The algorithm called DL-G-L1-adapt in the paper is tested here.

% BD 20.06.2024

%% ----------------------------------------
data_file_path = '../DATA_DEPENDENCY/dependency_';
data_file_names = {'annthyroid', 'breastw', 'cardio', 'cover', 'fault', ...
                   'glass', 'Hepatitis', 'Ionosphere', 'landsat', 'letter', ...
                   'Lymphography', 'magic.gamma', 'mammography', 'musk', 'pendigits', ...
                   'Pima', 'satellite', 'satimage', 'SpamBase', 'Stamps', ...
                   'thyroid', 'vertebral', 'vowels', 'Waveform', 'WBC', ...
                   'WDBC', 'Wilt', 'wine', 'WPBC', 'yeast' }; % full file list
data_file_names = {'wine', 'yeast'}; % this line is only for a first quick test
file_nr = int2str(randi(9999));
results_file_name = ['res_gauss_', file_nr, '.csv'];

n_tests = 5; % 5
train_data_ratio = 0.7;

dict_file_path = './save_dict_dependency/';
perm_file_path = './save_perm_dependency/';
load_dict_perm = 1;  % load initial trained dictionaries and train-test splits
save_err = 1;

overcompleteness_list = [2 2.5 3]; %[2]; % overcompleteness factor of the dictionary
lam_list = 1; %[0.001];
gam_list = 1 * lam_list;  % there should be some proportionality here
update_sigmas = 1;   % 0: constant sigmas; 1: automatically update sigmas
both_dl = 0;    % run both DL methods (with fixed sigmas and adapted sigmas); update_sigma must be 1

sigma_max = 0.1; % 0.1
sigma_min = 0.01; % 0.01
nit_gauss_repr = 100;
nit_gauss_dl = 100; %100;
nit_dl = 40;
s_dl = 2;
nit_fista = 50; %20  % number of FISTA iterations...
step_fista = 0.0001; % 0.001 % step in FISTA

if both_dl & ~update_sigmas
  error("update_sigmas cannot be 0")
end

pdf_sig = 3;     % distribution of sigmas
                 %   0:   uniform random
                 %   1:   min-max (50-50)
                 %   2:   min-max (80-20)
                 %   3:   uniform (linspace)
if update_sigmas
  nit_sig = 10;    % number of iterations at which sigmas are updated
  met_adapt = 1;   % adaptation method, 0: number of nonzero coefficients,
                   %                    1: sum of absolute values of coefficients
end

if both_dl
  method_names = {'AK-SVD + Gauss-L1', 'DL-Gauss-L1', 'DL-G-L1-adapt'};
else
  if update_sigmas
    method_names = {'AK-SVD + Gauss-L1', 'DL-G-L1-adapt'};
  else
    method_names = {'AK-SVD + Gauss-L1', 'DL-Gauss-L1'};
  end      
end

% ----------------------------------------
n_datasets = length(data_file_names);
n_methods = length(method_names);
n_params = length(overcompleteness_list)*length(lam_list);

% prepare file for results; write header
fid_res = fopen(results_file_name,'w');
fprintf(fid_res, "%s,%s,", 'Dataset', 'Method');
if save_err
  results_file_name_err = ['res_gauss_', file_nr, '_err.csv'];
  fid_res_err = fopen(results_file_name_err,'w');
  fprintf(fid_res_err, "%s,%s,", 'Dataset', 'Method');
end
for i_over = 1:length(overcompleteness_list)
  for i_s = 1 : length(lam_list)
    fprintf(fid_res, "%s,", ['o=',num2str(overcompleteness_list(i_over)), ' lam=', int2str(lam_list(i_s)), ' [err]']);
    fprintf(fid_res, "%s,", ['o=',num2str(overcompleteness_list(i_over)), ' lam=', int2str(lam_list(i_s)), ' [proba]']);
    fprintf(fid_res, "%s,", ['o=',num2str(overcompleteness_list(i_over)), ' lam=', int2str(lam_list(i_s)), ' [obj]']);
    if save_err
      fprintf(fid_res_err, "%s,", ['o=',num2str(overcompleteness_list(i_over)), ' lam=', int2str(lam_list(i_s)), ' [err]']);
    end
  end
end
fprintf(fid_res, "\n");
if save_err
  fprintf(fid_res_err, "\n");
end

for i=1:n_methods
  res_methods_err{i} = zeros(n_datasets,n_params);
  res_methods_proba{i} = zeros(n_datasets,n_params);
  res_methods_obj{i} = zeros(n_datasets,n_params);
  err_methods_test{i} = zeros(n_datasets,n_params);
  time_methods_train{i} = zeros(n_datasets,n_params);
  time_methods_test{i} = zeros(n_datasets,n_params);
end

for id = 1 : n_datasets
  fprintf("%s\n", data_file_names{id})
  filename = [data_file_path, data_file_names{id}, '.mat'];
  load(filename)
  X = double(Y);   % data are in X, labels are in y
  y = labels;
  clear Y, clear labels
  X = X';   % signals are on columns
  [m,N] = size(X);

  % preprocessing: Z-score normalization
  vm = mean(X,2);
  X = double(X) - repmat(vm,1,N);         % subtract mean
  sm = std(X,0,2);
  X = X ./ repmat(sm,1,N);
  
  col_normal = find(y==0);
  col_anomaly = find(y~=0);
  Nanomaly = length(col_anomaly);
  Nnormal = N - Nanomaly;
  anomaly_ratio = Nanomaly/N;
  
  Ntrain = round(train_data_ratio*N);
  Ntest = N - Ntrain;
  N_anomaly_train = round(anomaly_ratio*Ntrain);
  N_normal_train = Ntrain - N_anomaly_train;
  N_anomaly_test = Nanomaly - N_anomaly_train;
  N_normal_test = Nnormal - N_normal_train;
  
  for im = 1:n_methods
    res_roc_err{im} = zeros(n_tests, n_params);
    res_roc_proba{im} = zeros(n_tests, n_params);
    res_roc_obj{im} = zeros(n_tests, n_params);
    res_err{im} = zeros(n_tests, n_params);
    res_proba{im} =  zeros(n_tests, n_params);
    res_obj{im} = zeros(n_tests, n_params);
    time_train{im} = zeros(n_tests, n_params);
    time_test{im} = zeros(n_tests, n_params);
  end
  
  for it = 1 : n_tests
    % split train and test data, respecting proportion of anomaly
    if load_dict_perm
      perm_file_name = [perm_file_path, 'p_', data_file_names{id}, '_', num2str(it-1), '.mat'];
      load(perm_file_name);
    else
      p1 = randperm(Nnormal);
      p2 = randperm(Nanomaly);
    end
    Ytrain = X(:,[col_normal(p1(1:N_normal_train)); col_anomaly(p2(1:N_anomaly_train))]);
    Ytest = X(:,[col_normal(p1(N_normal_train+1:end)); col_anomaly(p2(N_anomaly_train+1:end))]);
    ytest = [zeros(N_normal_test,1); y(col_anomaly(p2(N_anomaly_train+1:end)))];
    N_anomaly_test_w = sum(ytest);
    
    i_param = 0;
      
    for i_over = 1:length(overcompleteness_list)
      n = round(m*overcompleteness_list(i_over));
      switch pdf_sig
        case 0      % uniform random
          sigma_unif = sigma_min + (sigma_max-sigma_min) * rand(n,1);
        case 1      % min-max (50-50)
          %sigma_unif = sigma_min + (sigma_max-sigma_min) * round(rand(n,1));
          h = round(0.5*n);
          sigma_unif = sigma_min + (sigma_max-sigma_min) * [zeros(h,1); ones(n-h,1)];
        case 2      % min-max (80-20)
          %sigma_unif = sigma_min + (sigma_max-sigma_min) * round(rand(n,1)-0.2);
          h = round(0.8*n);
          sigma_unif = sigma_min + (sigma_max-sigma_min) * [zeros(h,1); ones(n-h,1)];
        case 3      % uniform deterministic
          sigma_unif = linspace(sigma_min, sigma_max, n)';
      end
      D0 = normc(randn(m,n));
      
      for i_s = 1 : length(lam_list)
        i_param = i_param+1;
        lam = lam_list(i_s);
        gam = gam_list(i_s);
        sigma = sigma_unif;
            
        % training phase
        % standard DL (AK-SVD)
        tic
        if load_dict_perm
          dict_file_name = [dict_file_path, 'D1_', data_file_names{id}, '_o', num2str(overcompleteness_list(i_over)), ...
                      '_', num2str(it-1), '.mat'];
          load(dict_file_name);
        else
          D1 = DL(Ytrain, D0, s_dl, nit_dl);
        end
        t=toc;
        max(max(D1'*D1-eye(n)))
        time_train{1}(it,i_param) = t;
        time_train{2}(it,i_param) = t;  % D1 is needed also in DL-Gauss
        if both_dl
          time_train{3}(it,i_param) = t;
        end
        % Gauss DL, standard or adaptive
        tic
        if both_dl | ~update_sigmas
          [D2, X2] = dl_gauss_l1(Ytrain, D1, sigma, lam, gam, nit_gauss_dl, ...
                        nit_fista, step_fista);
        else
          [D2, X2, sig_adapt] = dl_gauss_l1_sig(Ytrain, D1, sigma, lam, gam, nit_gauss_dl, ...
                        nit_sig, met_adapt, nit_fista, step_fista);
        end
        t=toc;
        max(max(D2'*D2-eye(n)))
        time_train{2}(it,i_param) = time_train{2}(it,i_param) + t;

        % Gauss DL adaptive (if required)
        if both_dl
          tic
          [D3, X3, sig_adapt] = dl_gauss_l1_sig(Ytrain, D1, sigma, lam, gam, nit_gauss_dl, ...
                        nit_sig, met_adapt, nit_fista, step_fista);
          t=toc;
          max(max(D3'*D3-eye(n)))
          time_train{3}(it,i_param) = time_train{3}(it,i_param) + t;
        end
        
        % testing phase
        % AK-SVD + Gauss-OMP-L1
        tic
        errs = zeros(Ntest,1);
        probas = zeros(Ntest,1);
        objs = zeros(Ntest,1);
        for k = 1 : Ntest
          [x, Da, err, proba, obj] = gauss_l1_s(Ytest(:,k), D1, sigma, ...
                                                lam, gam, nit_gauss_repr, nit_fista, step_fista);
          errs(k) = err(end);
          % if any(isnan(errs))
          %   error("Nan !!!!!!!")
          % end
          probas(k) = proba(end);
          objs(k) = obj(end);
        end
        res_err{1}(it,i_param) = mean(errs);
        [~,~,~,roc_auc] = perfcurve(ytest, errs, 1);
        res_roc_err{1}(it,i_param) = roc_auc;
        [~,~,~,roc_auc] = perfcurve(ytest, probas, 1);
        res_roc_proba{1}(it,i_param) = roc_auc;
        [~,~,~,roc_auc] = perfcurve(ytest, objs, 1);
        res_roc_obj{1}(it,i_param) = roc_auc;
        t=toc;
        time_test{1}(it,i_param) = t;

        % DL-Gauss-L1, standard or adaptive
        tic
        errs = zeros(Ntest,1);
        probas = zeros(Ntest,1);
        objs = zeros(Ntest,1);
        for k = 1 : Ntest
          if both_dl | ~update_sigmas
            [x, Da, err, proba, obj] = gauss_l1_s(Ytest(:,k), D2, sigma, ...
                                                  lam, gam, nit_gauss_repr, nit_fista, step_fista);
          else
            [x, Da, err, proba, obj] = gauss_l1_s(Ytest(:,k), D2, sig_adapt, ...
                                                  lam, gam, nit_gauss_repr, nit_fista, step_fista);
          end
          % if any(isnan(errs))
          %   error("Nan !!!!!!!")
          % end
          errs(k) = err(end);
          probas(k) = proba(end);
          objs(k) = obj(end);
        end
        res_err{2}(it,i_param) = mean(errs);
        [~,~,~,roc_auc] = perfcurve(ytest, errs, 1);
        res_roc_err{2}(it,i_param) = roc_auc;
        [~,~,~,roc_auc] = perfcurve(ytest, probas, 1);
        res_roc_proba{2}(it,i_param) = roc_auc;
        [~,~,~,roc_auc] = perfcurve(ytest, objs, 1);
        res_roc_obj{2}(it,i_param) = roc_auc;
        t=toc;
        time_test{2}(it,i_param) = t;

        % Gauss DL adaptive (if required)
        if both_dl
          tic
          errs = zeros(Ntest,1);
          probas = zeros(Ntest,1);
          objs = zeros(Ntest,1);
          for k = 1 : Ntest
            [x, Da, err, proba, obj] = gauss_l1_s(Ytest(:,k), D3, sig_adapt, ...
                                                  lam, gam, nit_gauss_repr, nit_fista, step_fista);
            errs(k) = err(end);
            % if any(isnan(errs))
            %   error("Nan !!!!!!!")
            % end
            probas(k) = proba(end);
            objs(k) = obj(end);
          end
          res_err{3}(it,i_param) = mean(errs);
          [~,~,~,roc_auc] = perfcurve(ytest, errs, 1);
          res_roc_err{3}(it,i_param) = roc_auc;
          [~,~,~,roc_auc] = perfcurve(ytest, probas, 1);
          res_roc_proba{3}(it,i_param) = roc_auc;
          [~,~,~,roc_auc] = perfcurve(ytest, objs, 1);
          res_roc_obj{3}(it,i_param) = roc_auc;
          t=toc;
          time_test{3}(it,i_param) = t;
        end
      end
    end
  end
  %fprintf("ROC AUC: %f\n", mean(res_roc_auc(:,id)))
  for im = 1 : n_methods
    res_methods_err{im}(id,:) = mean(res_roc_err{im});
    res_methods_proba{im}(id,:) = mean(res_roc_proba{im});
    res_methods_obj{im}(id,:) = mean(res_roc_obj{im});
    err_methods_test{im}(id,:) = mean(res_err{im});
    time_methods_train{im}(id,:) = mean(time_train{im});
    time_methods_test{im}(id,:) = mean(time_test{im});
    fprintf("%25s[err]  : %4f\n", method_names{im}, max(res_methods_err{im}(id,:)))
    fprintf("%25s[proba]: %4f\n", method_names{im}, max(res_methods_proba{im}(id,:)))
    fprintf("%25s[obj]  : %4f\n", method_names{im}, max(res_methods_obj{im}(id,:)))
  end
  % write in file
  for im = 1 : n_methods
    fprintf(fid_res, "%s,%s,", data_file_names{id}, method_names{im});
    if save_err
      fprintf(fid_res_err, "%s,%s,", data_file_names{id}, method_names{im});
    end
    for ip = 1 : n_params
      fprintf(fid_res, "%f,", res_methods_err{im}(id,ip));
      fprintf(fid_res, "%f,", res_methods_proba{im}(id,ip));
      fprintf(fid_res, "%f,", res_methods_obj{im}(id,ip));
      if save_err
        fprintf(fid_res_err, "%f,", err_methods_test{im}(id,ip));
      end
    end
    fprintf(fid_res, "\n");
    if save_err
      fprintf(fid_res_err, "\n");
    end
  end
end

fclose(fid_res);
if save_err
  fclose(fid_res_err);
end

%fprintf("%25s  Time train: %f\n", method_names{1}, sum(time_methods_train{1}))
fprintf("%25s  Time train: ", method_names{1})
for ip = 1 : n_params
   fprintf("%f   ", sum(time_methods_train{1}(:,ip)))
end
fprintf("\n")
%fprintf("%25s  Time train: %f\n", method_names{2}, sum(time_methods_train{2}))
fprintf("%25s  Time train: %f\n", method_names{2})
for ip = 1 : n_params
   fprintf("%f   ", sum(time_methods_train{2}(:,ip)))
end
fprintf("\n")
if both_dl
  % fprintf("%25s  Time train: %f\n", method_names{3}, sum(time_methods_train{3}))
  fprintf("%25s  Time train: %f\n", method_names{3})
  for ip = 1 : n_params
     fprintf("%f   ", sum(time_methods_train{3}(:,ip)))
  end
  fprintf("\n")
end
%fprintf("%25s  Time test : %f\n", method_names{1}, sum(time_methods_test{1}))
fprintf("%25s  Time test : ", method_names{1})
for ip = 1 : n_params
   fprintf("%f   ", sum(time_methods_test{1}(:,ip)))
end
fprintf("\n")
%fprintf("%25s  Time test : %f\n", method_names{2}, sum(time_methods_test{2}))
fprintf("%25s  Time test : ", method_names{2})
for ip = 1 : n_params
   fprintf("%f   ", sum(time_methods_test{2}(:,ip)))
end
fprintf("\n")
if both_dl
  %fprintf("%25s  Time test : %f\n", method_names{3}, sum(time_methods_test{3}))
  fprintf("%25s  Time test : ", method_names{3})
  for ip = 1 : n_params
     fprintf("%f   ", sum(time_methods_test{3}(:,ip)))
  end
  fprintf("\n")
end

%fprintf("%25s  Mean error : %f\n", method_names{1}, mean(err_methods_test{1}))
fprintf("%25s  Mean error :", method_names{1})
for ip = 1 : n_params
   fprintf("%f   ", mean(err_methods_test{1}(:,ip)))
end
fprintf("\n")
%fprintf("%25s  Mean error : %f\n", method_names{2}, mean(err_methods_test{2}))
fprintf("%25s  Mean error: ", method_names{2})
for ip = 1 : n_params
   fprintf("%f   ", mean(err_methods_test{2}(:,ip)))
end
fprintf("\n")
if both_dl
  %fprintf("%25s  Mean error : %f\n", method_names{3}, mean(err_methods_test{3}))
  fprintf("%25s  Mean error: ", method_names{3})
  for ip = 1 : n_params
    fprintf("%f   ", mean(err_methods_test{3}(:,ip)))
  end
  fprintf("\n")
end
